import json
import logging
import os
import re
import uuid
from pathlib import Path
from types import SimpleNamespace

import yaml
from filelock import FileLock

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("pysondb")


class DataNotFoundError(Exception):
    """Exception raised if id not found.

    Attributes:
        data
    """

    def __init__(self, data):
        self.data = data


class IdNotFoundError(Exception):
    """Exception raised if id not found.

    Attributes:
        pk -- primary key / id
    """

    def __init__(self, pk):
        self.pk = pk


class SchemaError(Exception):
    """Exception raised for field/key errors."""

    def __init__(self, *args):
        self.args = args


class JsonDatabase:
    def __init__(self, filename, id_fieldname="id", log=False):
        a = Database().on(filename)
        sdx = Path(filename)

        self._id_fieldname = id_fieldname
        self.filename = filename
        self.lock = FileLock("{}.lock".format(self.filename))

        if not log:
            self._stop_log()

        logger.info("Database Filename: {0}".format(sdx))

    def _get_id(self):
        return int(str(uuid.uuid4().int)[:18])

    def _cast_id(self, pk):
        return int(pk)

    def _get_load_function(self):
        return json.load

    def _get_dump_function(self):
        return json.dump

    @staticmethod
    def _objectify(data):
        return json.loads(data, object_hook=lambda d: SimpleNamespace(**d))

    @staticmethod
    def _stop_log():
        logging.getLogger("pysondb").disabled = True
        logging.getLogger("filelock").disabled = True

    @property
    def id_fieldname(self):
        return self._id_fieldname

    def add(self, new_data):
        with self.lock:
            with open(self.filename, "r+") as db_file:
                db_data = self._get_load_function()(db_file)
                try:
                    if set(db_data["data"][0].keys()) == set(new_data.keys()).union(
                        [self.id_fieldname]
                    ):
                        new_data[self.id_fieldname] = self._get_id()
                        logger.debug("Append new data; {0}".format(new_data))
                        db_data["data"].append(new_data)
                        db_file.seek(0)
                        self._get_dump_function()(db_data, db_file)
                        return new_data[self.id_fieldname]
                    else:
                        raise SchemaError(
                            "db_keys: "
                            + ",".join(sorted(list(db_data["data"][0].keys()))),
                            "new_data_keys: "
                            + ",".join(
                                sorted(list(new_data.keys()) + [self.id_fieldname])
                            ),
                        )
                except IndexError:
                    new_data[self.id_fieldname] = self._get_id()
                    db_data["data"].append(new_data)
                    logger.debug("Add first data entry; {0}".format(new_data))
                    db_file.seek(0)
                    self._get_dump_function()(db_data, db_file)
                return new_data[self.id_fieldname]

    def addMany(self, new_data):
        with self.lock:
            with open(self.filename, "r+") as db_file:
                db_data = self._get_load_function()(db_file)
                try:
                    for d in new_data:
                        if set(db_data["data"][0].keys()) == set(d.keys()).union(
                            [self.id_fieldname]
                        ):
                            d[self.id_fieldname] = self._get_id()
                            db_data["data"].append(d)
                            db_file.seek(0)
                            self._get_dump_function()(db_data, db_file)
                except:
                    keys = list(new_data[0].keys())
                    for d in new_data:
                        if set(keys) == set(d.keys()):
                            d[self.id_fieldname] = self._get_id()
                            db_data["data"].append(d)
                            db_file.seek(0)
                            self._get_dump_function()(db_data, db_file)

    def getAll(self, objectify=False):
        with self.lock:
            with open(self.filename, "r", encoding="utf8") as db_file:
                db_data = self._get_load_function()(db_file)

            return (
                db_data["data"]
                if not objectify
                else self._objectify(json.dumps(db_data)).data
            )

    def get(self, num=1, objectify=False):
        with self.lock:
            try:
                with open(self.filename, "r", encoding="utf8") as db_file:
                    db_data = self._get_load_function()(db_file)
                if num <= len(db_data["data"]):
                    data = db_data["data"][0 : int(num)]
                    return (
                        data
                        if not objectify
                        else self._objectify(json.dumps({"data": data})).data
                    )
                else:
                    logger.info(
                        "The length you have given {} \n Length of the database items= {}".format(
                            num, len(db_data["data"])
                        )
                    )
                    return []
            except:
                return []

    def getBy(self, query, objectify=False):
        with self.lock:
            result = []
            with open(self.filename, "r") as db_file:
                db_data = self._get_load_function()(db_file)
                for d in db_data["data"]:
                    if all(x in d and d[x] == query[x] for x in query):
                        result.append(d)
            return (
                result
                if not objectify
                else self._objectify(json.dumps({"data": result})).data
            )

    def reSearch(self, key, _re, objectify=False):
        pattern = _re
        if not isinstance(_re, re.Pattern):
            pattern = re.compile(str(_re))

        items = []
        data = self.getAll()

        for d in data:
            for k in d.keys():
                if re.match(pattern, str(d[k])) and k == key:
                    items.append(d)
                    continue

        return (
            items
            if not objectify
            else self._objectify(json.dumps({"data": items})).data
        )

    def updateById(self, pk, new_data):
        with self.lock:
            with open(self.filename, "r+") as db_file:
                db_data = self._get_load_function()(db_file)
                result = []
                if set(new_data.keys()).issubset(db_data["data"][0].keys()):
                    for d in db_data["data"]:
                        if d[self.id_fieldname] == self._cast_id(pk):
                            d.update(new_data)
                            result.append(d)
                        else:
                            raise IdNotFoundError(pk)
                    db_data["data"] = result
                    db_file.seek(0)
                    db_file.truncate()
                    self._get_dump_function()(db_data, db_file)
                else:
                    raise SchemaError(
                        "db_keys: " + ",".join(sorted(db_data.keys())),
                        "new_keys: " + ",".join(sorted(new_data.keys())),
                    )

    def deleteById(self, pk):
        with self.lock:
            with open(self.filename, "r+") as db_file:
                db_data = self._get_load_function()(db_file)
                result = []
                found = False

                for d in db_data["data"]:
                    if d[self.id_fieldname] == self._cast_id(pk):
                        found = True
                    else:
                        result.append(d)

                if not found:
                    raise IdNotFoundError(pk)

                db_data["data"] = result
                db_file.seek(0)
                db_file.truncate()
                self._get_dump_function()(db_data, db_file)
            return True

    def update(self, db_dataset, new_dataset):
        with self.lock:
            with open(self.filename, "r+") as db_file:
                db_data = self._get_load_function()(db_file)
                result = []
                found = False
                if set(db_dataset.keys()).issubset(db_data["data"][0].keys()) and set(
                    new_dataset.keys()
                ).issubset(db_data["data"][0].keys()):
                    for d in db_data["data"]:
                        if all(x in d and d[x] == db_dataset[x] for x in db_dataset):
                            if set(new_dataset.keys()).issubset(
                                db_data["data"][0].keys()
                            ):
                                d.update(new_dataset)
                                result.append(d)
                                found = True
                        else:
                            result.append(d)

                    if not found:
                        raise DataNotFoundError(db_dataset)

                    db_data["data"] = result
                    db_file.seek(0)
                    db_file.truncate()
                    self._get_dump_function()(db_data, db_file)
                else:
                    raise SchemaError(
                        "db_dataset_keys: " + ",".join(sorted(list(db_dataset.keys()))),
                        "db_keys: " + ",".join(sorted(list(db_data["data"][0].keys()))),
                        "new_dataset_keys: "
                        + ",".join(sorted(list(new_dataset.keys()))),
                    )


class UuidDatabase(JsonDatabase):
    def _get_id(self):
        return str(uuid.uuid4())

    def _cast_id(self, pk):
        return pk


class YamlDatabase(JsonDatabase):
    def _get_load_function(self):
        return yaml.safe_load

    def _get_dump_function(self):
        return yaml.dump


class JsonUuidDatabase(UuidDatabase):
    pass


class YamlUuidDatabase(UuidDatabase, YamlDatabase):
    pass


EMPTY_DATA = {"data": []}


class Database:
    def create(self, filename, data):
        with open(filename, "w") as db_file:
            db_file.write(data)

    def on(self, filename, uuid=True, create=True):
        if filename.split(".")[-1:][0] == "json":
            if create and not os.path.exists(filename):
                self.create(filename, json.dumps(EMPTY_DATA))


getDb = JsonDatabase  # for legacy support.



import csv
import json
import os
import random

from typing import Optional

import fire
from beautifultable import BeautifulTable


def create_if_not_exist(file_name: str):
    """
    Checks for the existence of the provided JSON DB.
    If it does not, this will add {data:[]}.
    :param str file_name: The absolute path to the DB file
    """
    if not os.path.exists(file_name):
        with open(file_name, "w") as db_file:
            db = {"data": []}
            json.dump(db, db_file)
        print("Succesfully created {} in the directory.".format(file_name))


def display(file_name: str):
    table = BeautifulTable()
    with open(file_name) as jsondoc:
        data = json.load(jsondoc)
        real_data = data["data"]
        header = list(data["data"][0].keys())
        for all_data in real_data:
            table.rows.append(list(all_data.values()))
        table.columns.header = header
        print(table)


def delete(file_name: str):
    if os.path.exists(file_name):
        x = input("Do you want to remove the json file..(y/n)")
        if x in ["y", "Y"]:
            os.remove(file_name)
        else:
            print("Action terminated")
    else:
        print("The file does not exist")


def convert(csv_file, json_db):
    print("Reading data from {}".format(csv_file))
    arr = []
    with open(csv_file) as csvFile:
        csvReader = csv.DictReader(csvFile)
        for csvRow in csvReader:
            csvRow["id"] = random.randint(1000000000, 9999999999)
            arr.append(csvRow)
    print("Writing data into {}".format(json_db))
    x = {}
    x["data"] = arr
    with open(json_db, "w") as json_file:
        json.dump(x, json_file)
    print("Conversion successful")


def convert_db_to_csv(filename: str, targetcsv="converted.csv"):
    """
    Converts a JSON database to a csv.
    :param str filename: path of the target json file
    :param str targetcsv: path of the converted csv ,default : converted.csv
    """
    with open(filename, "r") as db:
        json_loaded = json.load(db)["data"]
        csv_file = open(targetcsv, "a")
        csv_writer = csv.writer(csv_file)
        header = json_loaded[0].keys()
        csv_writer.writerow(header)
        print("File converted and saving to ", targetcsv)
        for each in json_loaded:
            csv_writer.writerow(each.values())
        csv_file.close()


def merge(p_file: str, m_file: str, output_file: Optional[str] = None):
    """
    Merges two json DB with the same schema
    :param str p_file: The primary file
    :param str m_file: The file to combine with p_file
    :param str output_file: The name of the output file, default: p_file
    """

    def verify_file(file_data, refer_keys, filename):
        for d in file_data:
            temp_keys = list(set(d.keys()))
            temp_keys.sort()
            if not temp_keys == refer_keys:
                print(f"Irregularities in key names in database {filename!r}")
                quit()

    o_file = output_file or p_file
    with open(p_file, "r") as p, open(m_file) as m:
        try:
            p_data = json.load(p)["data"]
            m_data = json.load(m)["data"]

            # look up primary data: a reference to the first data entry
            lp_data = p_data[0]
            lm_data = m_data[0]

            # verify that all the entries in each DB have the same keys
            p_keys = list(set(lp_data.keys()))
            m_keys = list(set(lm_data.keys()))

            p_keys.sort()
            m_keys.sort()

            verify_file(p_data, p_keys, p_file)
            verify_file(m_data, m_keys, m_file)

        except KeyError:
            print("Oops, the DB's does not follow the required PysonDb schema.")
            quit()
        except IndexError:
            print("One of the Database is empty")
            quit()

    # merge the two DB together

    if len(lp_data) == len(lm_data):
        if all(i in lm_data.keys() for i in lp_data):
            temp_data = {"data": m_data + p_data}
            with open(o_file, "w") as f:
                json.dump(temp_data, f)
        else:
            print("The keys of the Database entries does not match")
    else:
        print("The number keys in DB entries does not match")


def main():
    fire.Fire(
        {
            "create": create_if_not_exist,
            "display": display,
            "delete": delete,
            "convert": convert,
            "converttocsv": convert_db_to_csv,
            "merge": merge,
        }
    )


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    author="Fredy Somy",
    author_email="fredysomy@gmail.com",
    python_requires=">=3",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    description="A Python JSON and/or YAML based lightweight Database.",
    install_requires=["fire", "beautifultable", "filelock", "pyaml"],
    license="MIT License",
    long_description=readme,
    long_description_content_type="text/markdown",
    version="1.0.3",
    keywords="pysondb,database,json,yaml",
    name="pysondb",
    packages=["pysondb"],
    entry_points={"console_scripts": ["pysondb=pysondb.cli:main"]},
    setup_requires=[],
    url="https://github.com/fredysomy/pysonDB",
    zip_safe=False,
)
